"""
Functions for working with numba.

Functionality includes:

- Functions for numba <-> riptable interop
- Convenience functions and decorators used when implementing numba-accelerated functions.
"""
__all__ = []
